@@ -6,7 +6,7 @@
 #                            | (__| |_| |  _ <| |___
 #                             \___|\___/|_| \_\_____|
 #
-# Copyright (C) 1998 - 2009, Daniel Stenberg, <daniel@haxx.se>, et al.
+# Copyright (C) 1998 - 2011, Daniel Stenberg, <daniel@haxx.se>, et al.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution. The terms
@@ -19,14 +19,14 @@
 # This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 # KIND, either express or implied.
 #
-# $Id: testcurl.pl,v 1.85 2009-07-14 15:36:12 gknauf Exp $
 ###########################################################################
 
 ###########################
 #  What is This Script?
 ###########################
 
-# testcurl.pl is the master script to use for automatic testing of CVS-curl.
+# testcurl.pl is the master script to use for automatic testing of curl
+# directly off its source repository.
 # This is written for the purpose of being run from a crontab job or similar
 # at a regular interval. The output is suitable to be mailed to
 # curl-autocompile@haxx.se to be dealt with automatically (make sure the
@@ -46,13 +46,15 @@
 # --extvercmd=[command]    Command to use for displaying version with cross compiles.
 # --mktarball=[command]    Command to run after completed test
 # --name=[name]            Set name to report as
-# --nocvsup                Don't update from CVS even though it is a CVS tree
+# --nocvsup                Don't pull from git even though it is a git tree
+# --nogitpull              Don't pull from git even though it is a git tree
 # --nobuildconf            Don't run buildconf
+# --noconfigure            Don't run configure
 # --runtestopts=[options]  Options to pass to runtests.pl
 # --setup=[file name]      File name to read setup from (deprecated)
 # --target=[your os]       Specify your target environment.
 #
-# if [curl-daily-name] is omitted, a 'curl' CVS directory is assumed.
+# if [curl-daily-name] is omitted, a 'curl' git directory is assumed.
 #
 
 use strict;
@@ -62,21 +64,28 @@ use Cwd;
 # Turn on warnings (equivalent to -w, which can't be used with /usr/bin/env)
 #BEGIN { $^W = 1; }
 
-use vars qw($version $fixed $infixed $CURLDIR $CVS $pwd $build $buildlog
-            $buildlogname $configurebuild $targetos $confsuffix $binext
+use vars qw($version $fixed $infixed $CURLDIR $git $pwd $build $buildlog
+            $buildlogname $configurebuild $targetos $confheader $binext
             $libext);
+
 use vars qw($name $email $desc $confopts $runtestopts $setupfile $mktarball
-            $extvercmd $nocvsup $nobuildconf $crosscompile $timestamp);
+            $extvercmd $nogitpull $nobuildconf $crosscompile
+            $timestamp $notes);
 
 # version of this script
-$version='$Revision: 1.85 $';
+$version='2011-12-27';
 $fixed=0;
 
-# Determine if we're running from CVS or a canned copy of curl,
+# Determine if we're running from git or a canned copy of curl,
 # or if we got a specific target option or setup file option.
 $CURLDIR="curl";
-$CVS=1;
+if (-f ".git/config") {
+  $CURLDIR = "./";
+}
+
+$git=1;
 $setupfile = 'setup';
+$configurebuild = 1;
 while ($ARGV[0]) {
   if ($ARGV[0] =~ /--target=/) {
     $targetos = (split(/=/, shift @ARGV))[1];
@@ -99,17 +108,22 @@ while ($ARGV[0]) {
   elsif ($ARGV[0] =~ /--desc=/) {
     $desc = (split(/=/, shift @ARGV))[1];
   }
-  elsif ($ARGV[0] =~ /--configure=/) {
-    $confopts = (split(/=/, shift @ARGV))[1];
+  elsif ($ARGV[0] =~ /--configure=(.*)/) {
+    $confopts = $1;
+    shift @ARGV;
   }
-  elsif ($ARGV[0] =~ /--nocvsup/) {
-    $nocvsup=1;
+  elsif (($ARGV[0] eq "--nocvsup") || ($ARGV[0] eq "--nogitpull")) {
+    $nogitpull=1;
     shift @ARGV;
   }
   elsif ($ARGV[0] =~ /--nobuildconf/) {
     $nobuildconf=1;
     shift @ARGV;
   }
+  elsif ($ARGV[0] =~ /--noconfigure/) {
+    $configurebuild=0;
+    shift @ARGV;
+  }
   elsif ($ARGV[0] =~ /--crosscompile/) {
     $crosscompile=1;
     shift @ARGV;
@@ -119,13 +133,12 @@ while ($ARGV[0]) {
   }
   else {
     $CURLDIR=shift @ARGV;
-    $CVS=0;
+    $git=0; # a given dir, assume not using git
   }
 }
 
 # Do the platform-specific stuff here
-$configurebuild = 1;
-$confsuffix = '';
+$confheader = 'curl_config.h';
 $binext = '';
 $libext = '.la'; # .la since both libcurl and libcares are made with libtool
 if ($^O eq 'MSWin32' || $targetos) {
@@ -133,7 +146,7 @@ if ($^O eq 'MSWin32' || $targetos) {
     # If no target defined on Win32 lets assume vc
     $targetos = 'vc';
   }
-  if ($targetos =~ /vc/ || $targetos =~ /borland/) {
+  if ($targetos =~ /vc/ || $targetos =~ /borland/ || $targetos =~ /watcom/) {
     $binext = '.exe';
     $libext = '.lib';
   }
@@ -155,15 +168,16 @@ if ($^O eq 'MSWin32' || $targetos) {
   }
 }
 
-if (($^O eq 'MSWin32') &&
-    ($targetos =~ /vc/ || $targetos =~ /mingw32/ || $targetos =~ /borland/)) {
+if (($^O eq 'MSWin32' || $^O eq 'msys') &&
+    ($targetos =~ /vc/ || $targetos =~ /mingw32/ ||
+     $targetos =~ /borland/ || $targetos =~ /watcom/)) {
 
   # Set these things only when building ON Windows and for Win32 platform.
   # FOR Windows since we might be cross-compiling on another system. Non-
-  # Windows builds still default to configure-style builds with no confsuffix.
+  # Windows builds still default to configure-style builds with curl_config.h.
 
   $configurebuild = 0;
-  $confsuffix = '-win32';
+  $confheader = 'config-win32.h';
 }
 
 $ENV{LC_ALL}="C" if (($ENV{LC_ALL}) && ($ENV{LC_ALL} !~ /^C$/));
@@ -294,6 +308,7 @@ if ($fixed < 4) {
     print F "email='$email'\n";
     print F "desc='$desc'\n";
     print F "confopts='$confopts'\n";
+    print F "notes='$notes'\n";
     print F "fixed='$fixed'\n";
     close(F);
 }
@@ -316,12 +331,14 @@ logit 'TRANSFER CONTROL ==== 1120 CHAR LINE' . $str1066os . 'LINE_END';
 logit "NAME = $name";
 logit "EMAIL = $email";
 logit "DESC = $desc";
+logit "NOTES = $notes";
 logit "CONFOPTS = $confopts";
 logit "CPPFLAGS = ".$ENV{CPPFLAGS};
 logit "CFLAGS = ".$ENV{CFLAGS};
 logit "LDFLAGS = ".$ENV{LDFLAGS};
 logit "CC = ".$ENV{CC};
 logit "MAKEFLAGS = ".$ENV{MAKEFLAGS};
+logit "ACLOCAL_FLAGS = ".$ENV{ACLOCAL_FLAGS};
 logit "PKG_CONFIG_PATH = ".$ENV{PKG_CONFIG_PATH};
 logit "target = ".$targetos;
 logit "version = $version"; # script version
@@ -334,15 +351,15 @@ $str1066os = undef;
 $pwd = getcwd();
 
 if (-d $CURLDIR) {
-  if ($CVS && -d "$CURLDIR/CVS") {
-    logit "$CURLDIR is verified to be a fine source dir";
+  if ($git && -d "$CURLDIR/.git") {
+    logit "$CURLDIR is verified to be a fine git source dir";
     # remove the generated sources to force them to be re-generated each
     # time we run this test
     unlink "$CURLDIR/src/hugehelp.c";
-  } elsif (!$CVS && -f "$CURLDIR/tests/testcurl.pl") {
+  } elsif (!$git && -f "$CURLDIR/tests/testcurl.pl") {
     logit "$CURLDIR is verified to be a fine daily source dir"
   } else {
-    mydie "$CURLDIR is not a daily source dir or checked out from CVS!"
+    mydie "$CURLDIR is not a daily source dir or checked out from git!"
   }
 }
 $build="build-$$";
@@ -368,55 +385,46 @@ if (-d $build) {
 # get in the curl source tree root
 chdir $CURLDIR;
 
-# Do the CVS thing, or not...
-if ($CVS) {
-
-  # this is a temporary fix to make things work again, remove later
-  logit "remove ares/aclocal.m4";
-  unlink "ares/aclocal.m4";
-
-  logit "update from CVS";
-  my $cvsstat;
-
-  sub cvsup() {
-    # update quietly to the latest CVS
-    if($nocvsup) {
-        logit "Skipping CVS update (--nocvsup)";
-        return 1;
+# Do the git thing, or not...
+if ($git) {
+  # update quietly to the latest git
+  if($nogitpull) {
+    logit "skipping git pull (--nogitpull)";
+  } else {
+    my $gitstat = 0;
+    my @commits;
+    logit "run git pull in curl";
+    system("git pull 2>&1");
+    $gitstat += $?;
+    logit "failed to update from curl git ($?), continue anyway" if ($?);
+    # get the last 5 commits for show (even if no pull was made)
+    @commits=`git log --pretty=oneline --abbrev-commit -5`;
+    logit "The most recent curl git commits:";
+    for (@commits) {
+      chomp ($_);
+      logit "  $_";
     }
-    else {
-        logit "run cvs up";
-        system("cvs -Q up -dP 2>&1");
+    if (-d "ares/.git") {
+      chdir "ares";
+      logit "run git pull in ares";
+      system("git pull 2>&1");
+      $gitstat += $?;
+      logit "failed to update from ares git ($?), continue anyway" if ($?);
+      # get the last 5 commits for show (even if no pull was made)
+      @commits=`git log --pretty=oneline --abbrev-commit -5`;
+      logit "The most recent ares git commits:";
+      for (@commits) {
+        chomp ($_);
+        logit "  $_";
+      }
+      chdir "$pwd/$CURLDIR";
     }
-
-    $cvsstat=$?;
-
-    # return !RETURNVALUE so that errors return 0 while goodness
-    # returns 1
-    return !$cvsstat;
-  }
-
-  my $att=0;
-  while (!cvsup()) {
-    $att++;
-    logit "failed CVS update attempt number $att.";
-    if ($att > 20) {
-      $cvsstat=111;
-      last; # get out of the loop
-    }
-    sleep 5;
-  }
-
-  if ($cvsstat != 0) {
-    mydie "failed to update from CVS ($cvsstat), exiting";
-  }
-  elsif (!$nocvsup) {
-    # Set timestamp to the UTC the CVS update took place.
-    $timestamp = scalar(gmtime)." UTC";
+    # Set timestamp to the UTC the git update took place.
+    $timestamp = scalar(gmtime)." UTC" if (!$gitstat);
   }
 
   if($nobuildconf) {
-      logit "told to not run buildconf";
+    logit "told to not run buildconf";
   }
   elsif ($configurebuild) {
     # remove possible left-overs from the past
@@ -424,13 +432,16 @@ if ($CVS) {
     unlink "autom4te.cache";
 
     # generate the build files
-    logit "invoke buildconf, but filter off aclocal underquoted definition warnings";
+    logit "invoke buildconf";
     open(F, "./buildconf 2>&1 |") or die;
     open(LOG, ">$buildlog") or die;
     while (<F>) {
-      next if /warning: underquoted definition of/;
-      print;
-      print LOG;
+      my $ll = $_;
+      # ignore messages pertaining to third party m4 files we don't care
+      next if ($ll =~ /aclocal\/gtk\.m4/);
+      next if ($ll =~ /aclocal\/gtkextra\.m4/);
+      print $ll;
+      print LOG $ll;
     }
     close(F);
     close(LOG);
@@ -447,14 +458,14 @@ if ($CVS) {
   }
 }
 
-# Set timestamp to the one in curlver.h if this isn't a CVS test build.
+# Set timestamp to the one in curlver.h if this isn't a git test build.
 if ((-f "include/curl/curlver.h") &&
     (open(F, "<include/curl/curlver.h"))) {
   while (<F>) {
     chomp;
     if ($_ =~ /^\#define\s+LIBCURL_TIMESTAMP\s+\"(.+)\".*$/) {
       my $stampstring = $1;
-      if ($stampstring !~ /CVS/) {
+      if ($stampstring !~ /DEV/) {
           $stampstring =~ s/\s+UTC//;
           $timestamp = $stampstring." UTC";
       }
@@ -499,6 +510,8 @@ if(!$make) {
 }
 # force to 'nmake' for VC builds
 $make = "nmake" if ($targetos =~ /vc/);
+# force to 'wmake' for Watcom builds
+$make = "wmake" if ($targetos =~ /watcom/);
 logit "going with $make as make";
 
 # change to build dir
@@ -566,14 +579,14 @@ else {
   mydie "no curlbuild.h created/found";
 }
 
-logit_spaced "display lib/curl_config$confsuffix.h";
-open(F, "lib/curl_config$confsuffix.h") or die "lib/curl_config$confsuffix.h: $!";
+logit_spaced "display lib/$confheader";
+open(F, "lib/$confheader") or die "lib/$confheader: $!";
 while (<F>) {
   print if /^ *#/;
 }
 close(F);
 
-if (grepfile("define USE_ARES", "lib/curl_config$confsuffix.h")) {
+if (grepfile("^#define USE_ARES", "lib/$confheader")) {
   print "\n";
   logit "setup to build ares";
 
@@ -602,8 +615,9 @@ if (grepfile("define USE_ARES", "lib/curl_config$confsuffix.h")) {
     mydie "no ares_build.h created/found";
   }
 
-  logit_spaced "display ares/ares_config$confsuffix.h";
-  if(open(F, "ares/ares_config$confsuffix.h")) {
+  $confheader =~ s/curl/ares/;
+  logit_spaced "display ares/$confheader";
+  if(open(F, "ares/$confheader")) {
       while (<F>) {
           print if /^ *#/;
       }
